use std::ops::Neg;

/// Marker trait, because certain operations only make sense on floating point types.
pub trait Float: Neg<Output=Self> {
    /// Calculate the absolute value. For primitive floats it just calls their `abs` function.
    fn abs(self) -> Self;

    /// Calculate the square root. For primitive floats it just calls their `sqrt` function.
    fn sqrt(self) -> Self;

    /// Calculate the sine.
    fn sin(self) -> Self;

    /// Calculate the cosine.
    fn cos(self) -> Self;

    /// Calculate the hyperbolic sine.
    fn sinh(self) -> Self;

    /// Calculate the hyperbolic cosine.
    fn cosh(self) -> Self;
}

impl Float for f32 {
    fn abs(self) -> Self {
        self.abs()
    }

    fn sqrt(self) -> Self {
        self.sqrt()
    }

    fn sin(self) -> Self {
        self.sin()
    }

    fn cos(self) -> Self {
        self.cos()
    }

    fn sinh(self) -> Self {
        self.sinh()
    }

    fn cosh(self) -> Self {
        self.cosh()
    }
}

impl Float for f64 {
    fn abs(self) -> Self {
        self.abs()
    }

    fn sqrt(self) -> Self {
        self.sqrt()
    }

    fn sin(self) -> Self {
        self.sin()
    }

    fn cos(self) -> Self {
        self.cos()
    }

    fn sinh(self) -> Self {
        self.sinh()
    }

    fn cosh(self) -> Self {
        self.cosh()
    }
}

#[cfg(test)]
mod tests {
    use crate::Float;

    fn is_float<T: Float>() -> bool {
        true
    }

    #[test]
    fn f32_is_a_float() {
        assert!(is_float::<f64>());
    }

    #[test]
    fn f64_is_a_float() {
        assert!(is_float::<f64>());
    }
}
