use std::fmt::Debug;
use std::ops::{Add, AddAssign, Div, DivAssign, Mul, MulAssign, Sub, SubAssign};

/// Trait representing all numeric primitives and `Complex`.
pub trait Numeric:
Sized + Copy + Clone + Debug +
Add<Output=Self> + AddAssign + Sub<Output=Self> + SubAssign +
Mul<Output=Self> + MulAssign + Div<Output=Self> + DivAssign +
PartialEq {
    /// Function returning the whole number in the requested type.
    ///
    /// # Safety
    /// The caller must make sure, that the given value can be represented in the target type.
    /// Otherwise behaviour is undefined.
    unsafe fn whole(value: u32) -> Self;
}

impl Numeric for f32 {
    unsafe fn whole(value: u32) -> Self {
        value as Self
    }
}

impl Numeric for f64 {
    unsafe fn whole(value: u32) -> Self {
        value as Self
    }
}

impl Numeric for i8 {
    unsafe fn whole(value: u32) -> Self {
        value as Self
    }
}

impl Numeric for i16 {
    unsafe fn whole(value: u32) -> Self {
        value as Self
    }
}

impl Numeric for i32 {
    unsafe fn whole(value: u32) -> Self {
        value as Self
    }
}

impl Numeric for i64 {
    unsafe fn whole(value: u32) -> Self {
        value as Self
    }
}

impl Numeric for i128 {
    unsafe fn whole(value: u32) -> Self {
        value as Self
    }
}

impl Numeric for isize {
    unsafe fn whole(value: u32) -> Self {
        value as Self
    }
}

impl Numeric for u8 {
    unsafe fn whole(value: u32) -> Self {
        value as Self
    }
}

impl Numeric for u16 {
    unsafe fn whole(value: u32) -> Self {
        value as Self
    }
}

impl Numeric for u32 {
    unsafe fn whole(value: u32) -> Self {
        value as Self
    }
}

impl Numeric for u64 {
    unsafe fn whole(value: u32) -> Self {
        value as Self
    }
}

impl Numeric for u128 {
    unsafe fn whole(value: u32) -> Self {
        value as Self
    }
}

impl Numeric for usize {
    unsafe fn whole(value: u32) -> Self {
        value as Self
    }
}

#[cfg(test)]
mod tests {
    use crate::Numeric;

    #[test]
    fn floats_are_numeric() {
        assert_eq!(unsafe { f32::whole(0) }, 0.0);
        assert_eq!(unsafe { f32::whole(1) }, 1.0);
        assert_eq!(unsafe { f64::whole(0) }, 0.0);
        assert_eq!(unsafe { f64::whole(1) }, 1.0);
    }

    #[test]
    fn integers_are_numeric() {
        // not sure yet, if I want this to be possible
        assert_eq!(unsafe { i8::whole(255) }, -1);

        assert_eq!(unsafe { i8::whole(0) }, 0);
        assert_eq!(unsafe { i8::whole(1) }, 1);
        assert_eq!(unsafe { i16::whole(0) }, 0);
        assert_eq!(unsafe { i16::whole(1) }, 1);
        assert_eq!(unsafe { i32::whole(0) }, 0);
        assert_eq!(unsafe { i32::whole(1) }, 1);
        assert_eq!(unsafe { i64::whole(0) }, 0);
        assert_eq!(unsafe { i64::whole(1) }, 1);
        assert_eq!(unsafe { i128::whole(0) }, 0);
        assert_eq!(unsafe { i128::whole(1) }, 1);
        assert_eq!(unsafe { isize::whole(0) }, 0);
        assert_eq!(unsafe { isize::whole(1) }, 1);

        assert_eq!(unsafe { u8::whole(0) }, 0);
        assert_eq!(unsafe { u8::whole(1) }, 1);
        assert_eq!(unsafe { u16::whole(0) }, 0);
        assert_eq!(unsafe { u16::whole(1) }, 1);
        assert_eq!(unsafe { u32::whole(0) }, 0);
        assert_eq!(unsafe { u32::whole(1) }, 1);
        assert_eq!(unsafe { u64::whole(0) }, 0);
        assert_eq!(unsafe { u64::whole(1) }, 1);
        assert_eq!(unsafe { u128::whole(0) }, 0);
        assert_eq!(unsafe { u128::whole(1) }, 1);
        assert_eq!(unsafe { usize::whole(0) }, 0);
        assert_eq!(unsafe { usize::whole(1) }, 1);
    }
}
